"""
Class "Airport" and Class "Path". 「机场」类和「航段」类。
"""

import json
from functools import cached_property, lru_cache, total_ordering
from urllib import request

from aviation_tools.fr24 import fr24, parse_json
from aviation_tools.point import Point


@total_ordering
class Airport(Point):
    """
    An Airport. 一座机场。

    Attributes:
        iata (str): IATA 3-digit Code 机场三字码
        lat (float): Latitude 纬度
        lng (float): Longitude 经度
        coord (float): (lat, lng) 经纬度表示
    """

    def __init__(self, iata: str):
        self.iata = iata
        Point.__init__(self, *self.position)

    @cached_property
    def position(self):
        return self.data['position']

    @cached_property
    def timezone(self):
        return self.data['timezone']

    @cached_property
    def name(self):
        return self.data['name']

    @cached_property
    def name_en(self):
        return self.data['name_en']

    @cached_property
    def city(self):
        return self.data['city']

    @cached_property
    def city_en(self):
        return self.data['city_en']

    @cached_property
    def icao(self):
        return self.data['ICAO']

    @cached_property
    def country(self):
        return self.data['country']

    @cached_property
    def province(self):
        return self.data['province'] if self.country == 'CN' else ''

    @cached_property
    def region(self):
        return self.data['region'] if self.country == 'CN' else ''

    @cached_property
    def geojson(self):
        return {
            "type": "Feature",
            "geometry": {
                "type": "Point",
                "coordinates": (self.lng.deg, self.lat.deg),
            },
            "properties": self.data
        }

    @cached_property
    def data(self):
        return Airport.raw_data()[self.iata]

    def __str__(self):
        return "{} {} {} ({}, {})".format(
            self.country, self.city, self.name, self.iata, self.icao)

    def __eq__(self, other):
        return self.iata == other.iata

    def __lt__(self, other):
        return self.iata < other.iata

    def __hash__(self):
        return self.iata.__hash__()

    @cached_property
    def neighbors(self):
        return [Airport(x) for x in fr24.get_destinations(self.iata)]

    @cached_property
    def fr24_properties(self):
        return fr24.get_airport(self.iata)

    @staticmethod
    @lru_cache()
    def raw_data():
        # https://liandlillc.github.io/airports_zh/airports.json
        return parse_json("data/airports.json")
