"""
Class "Angle". 「平面角」类。
"""

import math
import unittest
from functools import cached_property, total_ordering


@total_ordering
class Angle:
    """
    A planar Angle. 一个平面角。

    Attributes:
        raw_value (float): raw angle value. 原始值
        is_radian (bool): whether raw_value is
                          radian (True) or degree (False);
                          原始值是弧度（True）还是角度（False）
        deg (float): the angle in degree; 角度值
        rad (float): the angle in radian; 弧度值
        sin (float): the angle's sine; 正弦值
        cos (float): the angle's cosine; 余弦值
        tan (float): the angle's tangent; 正切值
    """

    def __init__(self, raw_value: float, is_radian: bool = False):
        raw_value = round(float(raw_value), 9)
        self.deg = math.degrees(raw_value) if is_radian else raw_value
        self.rad = raw_value if is_radian else math.radians(raw_value)

    @cached_property
    def sin(self):
        return math.sin(self.rad)

    @cached_property
    def cos(self):
        return math.cos(self.rad)

    @cached_property
    def tan(self):
        return math.tan(self.rad)

    def __neg__(self):
        return Angle(-self.deg)

    def __add__(self, other):
        return Angle(self.deg + float(other))

    def __sub__(self, other):
        return self + (-other)

    def __mul__(self, coefficient: float):
        return Angle(self.deg * coefficient)

    def __eq__(self, other) -> bool:
        return self.deg == float(other)

    def __lt__(self, other) -> bool:
        return self.deg < float(other)

    def __float__(self):
        return float(self.deg)

    def __str__(self):
        return str(self.deg)

    def standardize(self, left_bound: float = -180):
        """
        Standardize the Angle into a new Angle
        that satisfy periodic properties.

        将这个 Angle 标准化为一个符合周期性要求的新 Angle。

        Any circular Angle x can be expressed by
        x = 360°n + x', in which x' falls in [l, l + 360°).

        任意圆周角 x 可以表示为 x = 360°n + x'，
        其中 x' 位于 [l, l + 360°) 内。

        For longitude, l = -180°; for heading, l = 0.

        对经度而言 l = -180°；对航向角而言 l = 0。

        Args:
            left_bound (float): the left bound. 左界。

        Returns:
            float: The Angle falls in the given boundary, e.g.:
            位于边界内的角，例如：
            721° -> 1° (left_bound = 0).
            -181° -> 179° (left_bound = -180).
        """
        return Angle((self - left_bound).deg % 360) + left_bound

    @staticmethod
    def asin(value: float):
        return Angle(math.asin(value), True)

    @staticmethod
    def acos(value: float):
        return Angle(math.acos(value), True)

    @staticmethod
    def atan2(sin_value: float, cos_value: float = 1):
        return Angle(math.atan2(sin_value, cos_value), True)

    @staticmethod
    def asinh(value: float):
        return Angle(math.asinh(value), True)


if __name__ == "__main__":
    import unittest

    class TestDict(unittest.TestCase):
        def test_init(self):
            self.assertAlmostEqual(Angle(45).deg, 45.0)
            self.assertAlmostEqual(Angle(45).rad, math.pi / 4)

        def test_sin(self):
            self.assertAlmostEqual(Angle(30).sin, math.sin(math.pi / 6))

        def test_cos(self):
            self.assertAlmostEqual(Angle(60).cos, math.cos(math.pi / 3))

        def test_tan(self):
            self.assertAlmostEqual(Angle(45).tan, math.tan(math.pi / 4))

        def test_neg(self):
            self.assertEqual(-Angle(45).deg, (-Angle(45)).deg)

        def test_add(self):
            self.assertEqual((Angle(30) + Angle(45)).deg, Angle(75).deg)
            self.assertEqual((Angle(30) + 45).deg, Angle(75).deg)

        def test_sub(self):
            self.assertEqual((Angle(30) - Angle(45)).deg, Angle(-15).deg)
            self.assertEqual((Angle(30) - 45).deg, Angle(-15).deg)

        def test_mul(self):
            self.assertEqual((Angle(30) * 45).deg, Angle(30 * 45).deg)

        def test_eq(self):
            self.assertEqual(Angle(1080) == Angle(6 * math.pi, True), True)

        def test_lq(self):
            self.assertEqual(Angle(1080) < Angle(1100), True)

        def test_float(self):
            self.assertEqual(float(Angle(1080)), 1080.0)

        def test_str(self):
            self.assertEqual(str(Angle(1080)), '1080.0')

        def test_standardize(self):
            self.assertEqual(Angle(1080).standardize(0).deg, 0.0)
            self.assertEqual(Angle(540).standardize(0).deg, 180.0)
            self.assertEqual(Angle(540).standardize(-180).deg, -180.0)
            self.assertEqual(Angle(721).standardize(0).deg, 1.0)
            self.assertEqual(Angle(-181).standardize(-180).deg, 179.0)

        def test_asin(self):
            self.assertEqual(Angle.asin(1).deg, 90.0)
            self.assertEqual(Angle.asin(0).deg, 0.0)
            self.assertEqual(Angle.asin(-1).deg, -90.0)

        def test_acos(self):
            self.assertEqual(Angle.acos(1).deg, 0.0)
            self.assertEqual(Angle.acos(0).deg, 90.0)
            self.assertEqual(Angle.acos(-1).deg, 180.0)

        def test_atan2(self):
            self.assertEqual(Angle.atan2(1, 1).rad, math.atan2(1, 1))

        def test_asinh(self):
            self.assertEqual(Angle.asinh(0).rad, math.asinh(0))

    unittest.main()
