import json
import os
import re
import time
from functools import cached_property, lru_cache

import requests
from bs4 import BeautifulSoup

last_request = 0


def parse_json(filename):
    with open(os.path.join(os.path.split(__file__)[0], filename), encoding="utf-8") as w:
        return json.load(w)


class fr24:
    @staticmethod
    def get_html(url, headers={}):
        url_prefix = "https://www.flightradar24.com/data/"

        html = ""
        global last_request
        pux = 10 - (time.time() - last_request)
        if pux > 0:
            time.sleep(pux)
        last_request = time.time()

        config = parse_json("config/fr24.json")

        pre_headers = config["html_headers"]

        for i in headers:
            pre_headers[i] = headers[i]

        proxies = config["proxies"]
        full_url = url_prefix + url
        print("从 {url} 获取代码".format(url=full_url))
        html = (
            requests.get(
                full_url, headers=pre_headers, proxies=proxies,
            ).text.replace("</body>", "").replace("</html>", "") +
            "</body></html>")

        return BeautifulSoup(html, features="lxml")

    @staticmethod
    def get_json(url, params={}, user_headers={}, use_token=False):
        """
        从 flightradar24 获取一个 JSON 格式的文件

        参数:
            url: 需要获取的 URL
            params: 可选，默认为空，URL 参数
            user_headers: 可选，默认为空，请求头

        返回:
            使用 json.load 解析后的对象

        异常:
            暂无
        """
        global last_request
        pux = 10 - (time.time() - last_request)
        if pux > 0:
            time.sleep(pux)
        last_request = time.time()

        config = parse_json("config/fr24.json")

        TOKEN = config["json_token"]
        url_prefix = "https://api.flightradar24.com/common/v1/"
        if "token" not in params and use_token:
            params["token"] = TOKEN

        headers = config["json_headers"]

        for i in user_headers:
            headers[i] = user_headers[i]

        proxies = config["proxies"]
        full_url = url_prefix + url
        print("从 {url} 获取代码".format(url=full_url))
        u = requests.get(full_url, headers=headers,
                         params=params, proxies=proxies)

        return u.json()

    @staticmethod
    def item_handler(item, default_iata=None):
        """
        将一个完整的 flightradar24 机场/航空公司变成其 IATA / ICAO 代码

        参数:
            item: 一个含有 code 字段的字典

        返回:
            item.code.iata 或者 item.code.icao

        异常:
            暂无
        """

        if item and "code" in item:
            if "iata" in item["code"]:
                return item["code"]["iata"]
            if "icao" in item["code"]:
                return item["code"]["icao"]
        else:
            return default_iata

    @staticmethod
    def fr24_flight_converter(item, default_iata):
        """
        将一个完整的 flightradar24 飞行记录中的机场/航空公司变成其 IATA / ICAO 代码

        参数:
            item: 一段飞行记录

        返回:
            代码化后的飞行记录

        异常:
            暂无
        """

        if type(item) == list:
            return list(map(fr24.fr24_flight_converter, item))
        elif type(item) == dict:
            # 处理机场
            for i in item['airport']:
                item['airport'][i] = fr24.item_handler(
                    item['airport'][i], default_iata)
            # 处理航空公司
            item["airline"] = fr24.item_handler(item["airline"])
            item["owner"] = fr24.item_handler(item["owner"])
            item["aircraft"] = (
                {
                    "registration": item["aircraft"]["registration"],
                    "model": item["aircraft"]["model"],
                } if item["aircraft"] else {})

        return item

    @staticmethod
    def get_flight(flight_no, mode):
        """
        通过 flightradar24 获取一趟航班或一架飞机的飞行记录

        参数:
            flight_no: 航班号或者注册号
            mode: flight（航班号） 或 reg（注册号）

        返回:
            一个字典，info 键为该航班、飞机的信息；data 键为按日期倒序排列的飞行数据

        异常:
            暂无
        """

        if mode not in ["flight", "reg"]:
            raise KeyError('Valid mode should be either "flight" or "reg"')

        results = []
        timestamp = ""
        page = 1
        while True:
            params = {
                "query": flight_no,
                "fetchBy": mode,
                "page": page,
                "limit": 100,
                "pk": "",
                "timestamp": timestamp,
            }
            raw_html = fr24.get_json("flight/list.json",
                                     params)["result"]["response"]

            u = raw_html["data"]
            if u and len(u) > 0:
                results += fr24.fr24_flight_converter(u)
                timestamp = u[-1]["time"]["scheduled"]["departure"]
                page += 1
            else:
                break
            if not raw_html["page"]["more"]:
                break
        return results

    @staticmethod
    def get_flight_by_flight_no(flight_no):
        return fr24.get_flight(flight_no, "flight")

    @staticmethod
    def get_flight_by_aircraft_reg(aircraft_reg):
        return fr24.get_flight(aircraft_reg, "reg")

    @staticmethod
    def get_destinations(iata="", icao=""):
        url = "airports/{0}/routes" if icao == "" else "airlines/{0}-{1}/routes"
        url = url.format(iata.lower(), icao.lower())

        rv = []
        for i in fr24.get_html(url).find_all("script"):
            if "arrRoutes" in str(i):
                rv = json.loads(re.search(r"\[.*\]", str(i))[0])
                break

        if icao == "":
            return sorted([x['iata'] for x in rv if x['iata']])
        else:
            return rv

    @staticmethod
    def get_airport(iata):
        fr24.get_json("airport.json", {
            "code": iata.lower(),
            "page": 1,
            "limit": 100,
        })["result"]["response"]["airport"]["pluginData"]

    @staticmethod
    def get_destinations_of_airline(iata, icao):
        return fr24.get_destinations(iata, icao)

    @staticmethod
    def get_airline(iata, icao):
        iata, icao = iata.lower(), icao.lower()
        stri = icao if iata == "" else iata + "-" + icao

        p = fr24.get_html("airlines/{0}/fleet".format(stri))
        u = {}
        lists = p.find("dl", id="list-aircraft")
        if lists:
            for row in lists.find_all("tr"):
                cols = row.find_all("td")
                if len(cols) >= 2:
                    regLink, regType = cols[0:2]
                    regLink = regLink.find("a", class_="regLinks").text.strip()
                    regType = regType.text.strip()
                    if regType not in u:
                        u[regType] = []
                    u[regType].append(regLink)

        results = {"fleet": u}
        if iata == "":
            return results
        results["destinations"] = fr24.get_destinations_of_airline(iata, icao)
        return results


class fr24_Airport:
    def __init__(self, iata):
        self.iata = iata
        self.name_en = self.info['name']
        self.icao = self.info['code']['icao']
        self.country = self.info['position']['country']['code']
        self.city_en = self.info['position']['region']['city']

    @cached_property
    def info(self):
        print(self.iata, "调用了 FR24")

    @cached_property
    def position(self):
        return [
            self.info['position']['latitude'],
            self.info['position']['longitude']
        ]

    @cached_property
    def timezone(self):
        return self.info['timezone']['name']

    @cached_property
    def destinations(self):
        return fr24.get_destinations(self.iata)

    @cached_property
    def to_json(self):
        return {
            "IATA": self.iata,
            "ICAO": self.icao,
            "city_en": self.city_en,
            "country": self.country,
            "name_en": self.name_en,
            "position": self.position,
            "timezone": self.timezone
        }

    def __str__(self):
        return json.dumps(self.to_json, indent=2)
