"""
Class "Point". 「坐标点」类；
Class "Line". 「大圆线段」类。
"""

from functools import cached_property

from aviation_tools.angle import Angle
from aviation_tools.point import Point


class Line:
    """
    Great Circle Path between two points on earth.
    地球上两点之间的大圆航线。

    Attributes:
        src (Point): Source 起点
        dst (Point): Destination 终点
        angle (Angle): Central Angle 圆心角
        distance (float): Distance in km 以千米计的距离
        heading (Angle): Heading 航向角
    """

    def __init__(self, src: Point, dst: Point):
        self.src, self.dst = src, dst

    @cached_property
    def d_lng(self):
        return (self.dst.lng - self.src.lng).standardize(0)

    @cached_property
    def eastbound(self):
        return self.d_lng <= 180

    @cached_property
    def angle(self):
        return Angle.acos(
            self.src.lat.cos * self.dst.lat.cos * self.d_lng.cos
            + self.src.lat.sin * self.dst.lat.sin)

    @cached_property
    def distance(self):
        return self.angle.rad * Point.RADIUS

    @cached_property
    def heading(self):
        return Angle.atan2(
            self.dst.lat.cos * self.d_lng.sin,
            (self.src.lat.cos * self.dst.lat.sin
             - self.src.lat.sin * self.dst.lat.cos * self.d_lng.cos)
        ).standardize(0)

    @cached_property
    def azimuth(self):
        return Point(
            Angle.acos(self.src.lat.cos * self.heading.sin),
            Angle.atan2(self.heading.cos,
                        self.src.lat.sin * self.heading.sin
                        ) + self.src.lng)

    def great_circle_point(self, lng: float):
        """
        Given a longitude, calculate the point on the path.

        给定一个经度，计算路径上对应的点。

        Args:
            lng (float): a longitude 经度值

        Returns:
            Point: the corresponding point
        """
        return Point(Angle.atan2(
            Angle(self.azimuth.lng - lng).cos * self.azimuth.lat.sin,
            self.azimuth.lat.cos
        ), lng)

    @cached_property
    def points(self):
        """
        Draw the intermediate points using sampling.

        用取样法绘出中间点。
        """
        if not self.eastbound:
            return Line(self.dst, self.src).points
        src, dst = self.src.lng.deg, self.dst.lng.deg
        bound = 5
        if min(dst - src, 360 - dst + src) <= 2 * bound:
            return [self.src, self.dst]
        src = src // bound * bound + bound
        dst = dst // bound * bound
        rv = [self.src]
        i = Angle(src + bound)
        while True:
            rv.append(self.great_circle_point(i))
            if i == dst:
                break
            i = (i + bound).standardize()
        return rv + [self.dst]


if __name__ == "__main__":
    import unittest

    class TestDict(unittest.TestCase):
        def test_init(self):
            s, r = Point(0, 0), Point(0, 180)
            self.assertEqual(Line(s, r).dst, r)
            self.assertEqual(Line(s, r).src, s)

        def test_dlng(self):
            s, r = Point(0, 0), Point(0, 179)
            self.assertEqual(Line(s, r).d_lng.deg, 179)
            self.assertEqual(Line(r, s).d_lng.deg, 181)

        def test_eastbound(self):
            s, r = Point(0, 0), Point(0, 179)
            self.assertEqual(Line(s, r).eastbound, True)
            self.assertEqual(Line(r, s).eastbound, False)

        def test_angle(self):
            s, r, t = Point(0, 0), Point(0, 179), Point(90, 0)
            self.assertAlmostEqual(Line(s, r).angle.deg, 179)
            self.assertAlmostEqual(Line(r, s).angle.deg, 179)
            self.assertAlmostEqual(Line(s, t).angle.deg, 90)
            self.assertAlmostEqual(Line(r, t).angle.deg, 90)

        def test_distance(self):
            s, r = Point(0, 0), Point(0, 179)
            l = Line(s, r)
            self.assertEqual(l.distance, l.angle.rad * Point.RADIUS)

        def test_heading(self):
            s, r, t = Point(0, 0), Point(0, 179), Point(90, 0)
            self.assertAlmostEqual(Line(s, r).heading.deg, 90)
            self.assertAlmostEqual(Line(r, s).heading.deg, 270)
            self.assertAlmostEqual(Line(s, t).heading.deg, 0)
            self.assertAlmostEqual(Line(r, t).heading.deg, 0)
    unittest.main()
