"""
Class "Point". 「坐标点」类；
Class "Line". 「大圆线段」类。
"""

from functools import cached_property, total_ordering

from aviation_tools.angle import Angle


@total_ordering
class Point:
    """
    A point on earth. 地球上一点。

    Attributes:
        lat (float): Latitude 纬度
        lng (float): Longitude 经度
        coord (float): (lat, lng) 经纬度表示
    """

    RADIUS = 6371.009

    def __init__(self, lat, lng, is_radian: bool = False):
        assert (lat <= 90 and lat >= -90 and lng <= 180 and lng > -180)
        self.lat = Angle(lat, is_radian)
        self.lng = Angle(lng, is_radian).standardize()

    @cached_property
    def coord(self):
        return (self.lat.deg, self.lng.deg)

    @cached_property
    def geojson(self):
        return {"type": "Point", "coordinates": (self.lng.deg, self.lat.deg)}

    def __str__(self):
        return "({} {})".format(str(self.lat), str(self.lng))

    def __eq__(self, other) -> bool:
        return self.lat == other.lat and self.lng == other.lng

    def __lt__(self, other) -> bool:
        return (self.lat < other.lat) or (self.lat == other.lat and self.lng < other.lng)

    def destination(self, heading: float, distance: float):
        """
        Given heading and distance, calculate the destination's coordinates.

        给定航向角 heading 和距离 distance，计算终点的经纬度坐标。

        Args:
            heading (float): in degree 以度数表示的方向角。
            distance (float): in kilometers 以公里表示的距离。

        Returns:
            Point: the destination
        """

        heading_angle = Angle(heading)
        central_angle = Angle(distance / Point.RADIUS, True)
        temp = self.lat.cos * central_angle.sin
        lat2 = Angle.asin(temp * heading_angle.cos +
                          self.lat.sin * central_angle.cos)
        lng2 = self.lng + Angle.atan2(temp * heading_angle.sin,
                                      central_angle.cos - self.lat.sin * lat2.sin)
        return Point(lat2, lng2.standardize())

    def range(self, distance: float):
        """
        Given distance, generate a series of destination that form a circle.

        给定距离 distance，计算距离形成的圆。

        Args:
            distance (float): in kilometers 以公里表示的距离。

        Generates:
            Point: a point 圆上一点。
        """
        return [self.destination(heading, distance) for heading in range(-90, 271)]


if __name__ == "__main__":
    import unittest

    class TestDict(unittest.TestCase):
        def test_destination(self):
            s, r = Point(0, 0), Point.RADIUS
            self.assertAlmostEqual(s.destination(0, r).lat.rad, 1)
            self.assertAlmostEqual(s.destination(0, r).lng.deg, 0)
            self.assertAlmostEqual(s.destination(90, r).lat.deg, 0)
            self.assertAlmostEqual(s.destination(90, r).lng.rad, 1)
            self.assertAlmostEqual(s.destination(180, r).lat.rad, -1)
            self.assertAlmostEqual(s.destination(180, r).lng.deg, 0)
            self.assertAlmostEqual(s.destination(270, r).lat.deg, 0)
            self.assertAlmostEqual(s.destination(270, r).lng.rad, -1)

    unittest.main()
